// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "TicTacToeGame.hpp"
#include "game/DawnGame.hpp"
#include "scene/components/example/ExampleSpin.hpp"
#include "scene/components/physics/3d/CubeCollider.hpp"

using namespace Dawn;

TicTacToeGame::TicTacToeGame(SceneItem *item) :
  SceneItemComponent(item),
  winner(TIC_TAC_TOE_EMPTY),
  nextMove(TIC_TAC_TOE_NOUGHT)
{}

void TicTacToeGame::onStart() {
  // Map tiles by tile number = tile
  auto ts = getScene()->findComponents<TicTacToeTile>();
  auto itTiles = ts.begin();
  while(itTiles != ts.end()) {
    this->tiles[(*itTiles)->tile] = *itTiles;
    ++itTiles;
  }

  useEffect([&]{
    if(nextMove != TIC_TAC_TOE_CROSS) return;
    std::cout << "AI Move" << std::endl;
  }, nextMove);

  useEffect([&]{
    if(winner == TIC_TAC_TOE_NOUGHT) return;
    std::cout << "Winner is " << winner << std::endl;
  }, winner);

  useEvent([&](float_t delta) {
    // Only allow player input if it's their turn.
    if(nextMove != TIC_TAC_TOE_NOUGHT) return;

    // Get mouse in screen space.
    auto mouse = getGame()->inputManager.getAxis2D(INPUT_BIND_MOUSE_X, INPUT_BIND_MOUSE_Y);
    mouse *= 2.0f;
    mouse -= glm::vec2(1, 1);

    Camera *camera = getScene()->findComponent<Camera>();
    if(camera == nullptr) return;

    struct Ray3D ray;
    ray.origin = camera->transform->getWorldPosition();
    ray.direction = camera->getRayDirectionFromScreenSpace(mouse);
    

    // Find the hovered tile (if any)
    TicTacToeTile *hovered = nullptr;
    auto results = getPhysics()->raycast3DAll(ray);
    auto itResult = results.begin();
    while(itResult != results.end()) {
      auto result = *itResult;
      auto tile = result.collider->item->getComponent<TicTacToeTile>();
      if(tile == nullptr) {
        ++itResult;
        continue;
      }
      
      hovered = tile;
      break;
    }

    // Now update the state of each tile, also get the state while we are at it
    std::map<uint8_t, enum TicTacToeTileState> tileMap;

    auto itTiles = tiles.begin();
    while(itTiles != tiles.end()) {
      auto t = itTiles->second;

      if(t == hovered) {
        if(t->tileState == TIC_TAC_TOE_EMPTY && getGame()->inputManager.isPressed(INPUT_BIND_MOUSE_CLICK)) {
          t->tileState = nextMove;
          nextMove = nextMove == TIC_TAC_TOE_NOUGHT ? TIC_TAC_TOE_CROSS : TIC_TAC_TOE_NOUGHT;
        } else if(t->tileState == TIC_TAC_TOE_EMPTY) {
          t->hovered = true;
        }
      } else {
        t->hovered = false;
      }
      tileMap[itTiles->second->tile] = itTiles->second->tileState;
      ++itTiles;
    }

    // Determine winner
    std::vector<uint8_t> winningCombo;
    winner = ticTacToeDetermineWinner(tileMap, &winningCombo);
  }, getScene()->eventSceneUpdate);
}